import importlib

class PluginEntry(object):
    """A Plugin entry point

    The entry-point is not loaded/imported on creation.
    Use the method `get()` to import the module and get the attribute.
    """

    class Sentinel(object):
        pass

    # indicate the entry-point object is not loaded yet
    NOT_LOADED = Sentinel()

    def __init__(self, category, name, location):
        """
        :param category str: plugin category name
        :param name str: plugin name (as used by doit)
        :param location str: python object location as <module>:<attr>
        """
        self.obj = self.NOT_LOADED
        self.category = category
        self.name = name
        self.location = location

    def __repr__(self):
        return "PluginEntry('{}', '{}', '{}')".format(
            self.category, self.name, self.location)

    def get(self):
        """return obj, get from cache or load"""
        if self.obj is self.NOT_LOADED:
            self.obj = self.load()
        return self.obj

    def load(self):
        """load/import reference to obj from named module/obj"""
        module_name, obj_name = self.location.split(':')
        try:
            module = importlib.import_module(module_name)
        except ImportError:
            raise Exception('Plugin {} module `{}` not found.'.format(
                self.category, module_name))
        try:
            obj = getattr(module, obj_name)
        except AttributeError:
            raise Exception('Plugin {}:{} module `{}` has no {}.'.format(
                self.category, self.name, module_name, obj_name))
        return obj


class PluginDict(dict):
    """A dict where item values *might* be a PluginEntry or a direct reference to class/obj.

    Values should not be accessed directly, use `get_plugin()` to make sure plugin is loaded.

    Typically, one dict is created for each kind of plugin.
    doit supports 4 categories:
     - COMMAND
     - LOADER
     - BACKEND
     - REPORTER
    """

    entry_point_prefix = 'doit'

    def add_plugins(self, cfg_data, category):
        """read all items from a ConfigParser section containing plugins & entry-points.

        Plugins from entry-point have higher priority
        """
        self.update(self._from_ini(cfg_data, category))
        self.update(self._from_entry_points(category))

    def _from_ini(self, cfg_data, category):
        """plugins from INI file

        INI `section` names map exactly to plugin `category`.
        """
        result = {}
        if category in cfg_data:
            for name, location in cfg_data[category].items():
                result[name] = PluginEntry(category, name, location)
        return result

    def _from_entry_points(self, category):
        """get all plugins from setuptools entry_points"""
        result = {}
        try:
            import pkg_resources
            group = f"{self.entry_point_prefix}.{category}"
            for point in pkg_resources.iter_entry_points(group=group):
                name = point.name
                location = "{}:{}".format(point.module_name, point.attrs[0])
                result[name] = PluginEntry(category, name, location)
        except ImportError: # pragma: no cover
            pass  # ignore, if setuptools is not installed
        return result


    def get_plugin(self, key):
        """load and return a single plugin"""
        val = self[key]
        if isinstance(val, PluginEntry):
            val.name = key # overwrite obj name attribute
            return val.get()
        else:
            return val

    def to_dict(self):
        """return a standard dict with all plugins values loaded (no PluginEntry)"""
        return {k: self.get_plugin(k) for k in self.keys()}



"""starts a long-running process that watches the file system and
automatically execute tasks when file dependencies change"""

import os
import time
import sys
from multiprocessing import Process
from subprocess import call

from .exceptions import InvalidCommand
from .cmdparse import CmdParse
from .filewatch import FileModifyWatcher
from .cmd_base import tasks_and_deps_iter
from .cmd_base import DoitCmdBase
from .cmd_run import opt_verbosity, Run

opt_reporter = {
    'name':'reporter',
    'short': None,
    'long': None,
    'type':str,
    'default': 'executed-only',
}

opt_success = {
    'name':'success_callback',
    'short': None,
    'long': 'success',
    'type':str,
    'default': '',
}

opt_failure = {
    'name':'failure_callback',
    'short': None,
    'long': 'failure',
    'type':str,
    'default': '',
}


class Auto(DoitCmdBase):
    """the main process will never load tasks,
    delegates execution to a forked process.

    python caches imported modules,
    but using different process we can have dependencies on python
    modules making sure the newest module will be used.
    """

    doc_purpose = "automatically execute tasks when a dependency changes"
    doc_usage = "[TASK ...]"
    doc_description = None
    execute_tasks = True

    cmd_options = (opt_verbosity, opt_reporter, opt_success, opt_failure)

    @staticmethod
    def _find_file_deps(tasks, sel_tasks):
        """find all file deps
        @param tasks (dict)
        @param sel_tasks(list - str)
        """
        deps = set()
        for task in tasks_and_deps_iter(tasks, sel_tasks):
            deps.update(task.file_dep)
            deps.update(task.watch)
        return deps


    @staticmethod
    def _dep_changed(watch_files, started, targets):
        """check if watched files was modified since execution started"""
        for watched in watch_files:
            # assume that changes to targets were done by doit itself
            if watched in targets:
                continue
            if os.stat(watched).st_mtime > started:
                return True
        return False


    @staticmethod
    def _run_callback(result, success_callback, failure_callback):
        '''run callback if any after task execution'''
        if result == 0:
            if success_callback:
                call(success_callback, shell=True)
        else:
            if failure_callback:
                call(failure_callback, shell=True)


    def run_watch(self, params, args):
        """Run tasks and wait for file system event

        This method is executed in a forked process.
        The process is terminated after a single event.
        """
        started = time.time()

        # execute tasks using Run Command
        arun = Run(task_loader=self.loader)
        params.add_defaults(CmdParse(arun.get_options()).parse([])[0])
        try:
            result = arun.execute(params, args)
        # ??? actually tested but coverage doesnt get it...
        except InvalidCommand as err: # pragma: no cover
            sys.stderr.write("ERROR: %s\n" % str(err))
            sys.exit(3)

        # user custom callbacks for result
        self._run_callback(result,
                           params.pop('success_callback', None),
                           params.pop('failure_callback', None))

        # get list of files to watch on file system
        watch_files = self._find_file_deps(arun.control.tasks,
                                           arun.control.selected_tasks)

        # Check for timestamp changes since run started,
        # if change, restart straight away
        if not self._dep_changed(watch_files, started, arun.control.targets):
            # set event handler. just terminate process.
            class DoitAutoRun(FileModifyWatcher):
                def handle_event(self, event):
                    # print("FS EVENT -> {}".format(event))
                    sys.exit(result)
            file_watcher = DoitAutoRun(watch_files)
            # kick start watching process
            file_watcher.loop()


    def execute(self, params, args):
        """loop executing tasks until process is interrupted"""
        while True:
            try:
                proc = Process(target=self.run_watch, args=(params, args))
                proc.start()
                proc.join()
                # if error on given command line, terminate.
                if proc.exitcode == 3:
                    return 3
            except KeyboardInterrupt:
                return 0



from .cmd_base import DoitCmdBase, check_tasks_exist, subtasks_iter

opt_listall = {
    'name': 'subtasks',
    'short':'',
    'long': 'all',
    'type': bool,
    'default': False,
    'help': "list include all sub-tasks from dodo file"
    }

opt_list_quiet = {
    'name': 'quiet',
    'short': 'q',
    'long': 'quiet',
    'type': bool,
    'default': False,
    'help': 'print just task name (less verbose than default)'
    }

opt_list_status = {
    'name': 'status',
    'short': 's',
    'long': 'status',
    'type': bool,
    'default': False,
    'help': 'print task status (R)un, (U)p-to-date, (I)gnored'
    }

opt_list_private = {
    'name': 'private',
    'short': 'p',
    'long': 'private',
    'type': bool,
    'default': False,
    'help': "print private tasks (start with '_')"
    }

opt_list_dependencies = {
    'name': 'list_deps',
    'short': '',
    'long': 'deps',
    'type': bool,
    'default': False,
    'help': ("print list of dependencies "
             "(file dependencies only)")
    }

opt_template = {
    'name': 'template',
    'short': '',
    'long': 'template',
    'type': str,
    'default': None,
    'help': "display entries with template"
    }

opt_sort = {
    'name': 'sort',
    'short': '',
    'long': 'sort',
    'type': str,
    'choices': [('name', 'sort by task name'),
                ('definition', 'list tasks in the order they were defined')],
    'default': 'name',
    'help': ("choose the manner in which the task list is sorted. "
             "[default: %(default)s]")
    }


class List(DoitCmdBase):
    doc_purpose = "list tasks from dodo file"
    doc_usage = "[TASK ...]"
    doc_description = None

    cmd_options = (opt_listall, opt_list_quiet, opt_list_status,
                   opt_list_private, opt_list_dependencies, opt_template,
                   opt_sort)


    STATUS_MAP = {'ignore': 'I', 'up-to-date': 'U', 'run': 'R', 'error': 'E'}


    def _print_task(self, template, task, status, list_deps, tasks):
        """print a single task"""
        line_data = {'name': task.name, 'doc':task.doc}
        # FIXME group task status is never up-to-date
        if status:
            # FIXME: 'ignore' handling is ugly
            if self.dep_manager.status_is_ignore(task):
                task_status = 'ignore'
            else:
                task_status = self.dep_manager.get_status(task, tasks).status
            line_data['status'] = self.STATUS_MAP[task_status]

        self.outstream.write(template.format(**line_data))

        # print dependencies
        if list_deps:
            for dep in task.file_dep:
                self.outstream.write(" -  %s\n" % dep)
            self.outstream.write("\n")

    @staticmethod
    def _list_filtered(tasks, filter_tasks, include_subtasks):
        """return list of task based on selected 'filter_tasks' """
        check_tasks_exist(tasks, filter_tasks)

        # get task by name
        print_list = []
        for name in filter_tasks:
            task = tasks[name]
            print_list.append(task)
            if include_subtasks:
                print_list.extend(subtasks_iter(tasks, task))
        return print_list


    def _list_all(self, include_subtasks):
        """list of tasks"""
        print_list = []
        for task in self.task_list:
            if (not include_subtasks) and task.subtask_of:
                continue
            print_list.append(task)
        return print_list


    def _execute(self, subtasks=False, quiet=True, status=False, private=False,
                 list_deps=False, template=None, sort='name', pos_args=None):
        """List task generators"""
        filter_tasks = pos_args
        # dict of all tasks
        tasks = dict([(t.name, t) for t in self.task_list])

        if filter_tasks:
            # list only tasks passed on command line
            print_list = self._list_filtered(tasks, filter_tasks, subtasks)
        else:
            print_list = self._list_all(subtasks)

        # exclude private tasks
        if not private:
            print_list = [t for t in print_list if not t.name.startswith('_')]

        # set template
        if template is None:
            max_name_len = 0
            if print_list:
                max_name_len = max(len(t.name) for t in print_list)

            template = '{name:<' + str(max_name_len + 3) + '}'
            if not quiet:
                template += '{doc}'
            if status:
                template = '{status} ' + template
        template += '\n'

        # sort list of tasks
        if sort == 'name':
            print_list = sorted(print_list)
        elif sort == 'definition':
            pass  # task list is already sorted in order of definition

        # print list of tasks
        for task in print_list:
            self._print_task(template, task, status, list_deps, tasks)
        return 0
